/*========================================================================
  Function Viewer: visualizer module
  ---------------
  This program is distributed under the GPL license.
==========================================================================*/

#include "Visualizer.hh"

#include <cstdio>
#include <cmath>

template<typename CBFunc>
void Visualizer::createButton(const char* name, CBFunc cb,
                              GtkWidget* buttonBar)
{
    GtkWidget* button = gtk_button_new_with_mnemonic(name);
    g_signal_connect(GTK_OBJECT(button), "clicked",
                     G_CALLBACK(cb), this);
    gtk_box_pack_start(GTK_BOX(buttonBar), button, FALSE, FALSE, 0);
}

//---------------------------------------------------------------------
// Constructor and configuration
//---------------------------------------------------------------------
Visualizer::Visualizer(GtkWidget* parent, GtkWidget* toolbar):
    drawingArea(gtk_drawing_area_new()),
    pixMap(0),
    drawingGC(gdk_gc_new(gtk_widget_get_root_window(parent))),
    colorMap(gdk_colormap_get_system()),
    pLayout(gtk_widget_create_pango_layout(parent, "0")),
    centerX(0), centerY(0),
    horizontalSize(10), aspectRatio(1)
{
    // Enable exposure, mouse button press and mouse motion (when button
    // is pressed) for the drawing area widget:
    gtk_widget_set_events(drawingArea,
                          GDK_EXPOSURE_MASK |
                          GDK_BUTTON_PRESS_MASK |
                          GDK_BUTTON_MOTION_MASK);

    // Connect the necessary signals to the drawing area widget:
    g_signal_connect(GTK_OBJECT(drawingArea), "expose_event",
                     G_CALLBACK(exposeCallback), this);
    g_signal_connect(GTK_OBJECT(drawingArea), "configure_event",
                     G_CALLBACK(configureCallback), this);
    g_signal_connect(GTK_OBJECT(drawingArea), "button-press-event",
                     G_CALLBACK(mouseClickCallback), this);
    g_signal_connect(GTK_OBJECT(drawingArea), "motion-notify-event",
                     G_CALLBACK(mouseMotionCallback), this);

    // Add buttons to the toolbar:
    createButton("Reset _scroll", resetScrollCallback, toolbar);
    createButton("Reset _zoom", resetZoomCallback, toolbar);
    createButton("Reset _y-scale", resetAspectRatioCallback, toolbar);
}

Visualizer::~Visualizer()
{
    g_object_unref(pLayout);
}

void Visualizer::addFunction(Function* f)
{
    functions.push_back(f);
    f->registerCallback(this);
}

void Visualizer::functionIsBeingDestroyed(Function* f)
{
    std::list<Function*>::iterator iter = functions.begin();
    while(iter != functions.end() && *iter != f) ++iter;
    functions.erase(iter);
}

void Visualizer::functionHasChanged()
{
    draw();
}



//---------------------------------------------------------------------
// Drawing functionality
//---------------------------------------------------------------------
inline void Visualizer::setColor(unsigned red, unsigned green, unsigned blue)
{
    GdkColor col = { 0, red, green, blue };
    gdk_color_alloc(colorMap, &col);
    gdk_gc_set_foreground(drawingGC, &col);
}

inline void Visualizer::refreshArea(int x, int y, int w, int h)
{
    gdk_draw_pixmap(drawingArea->window,
                    drawingArea->style->fg_gc[GTK_WIDGET_STATE(drawingArea)],
                    pixMap, x, y, x, y, w, h);
}

inline void Visualizer::horizontalLine(int y)
{
    gdk_draw_line(pixMap, drawingGC, 0, y, width-1, y);
}

inline void Visualizer::verticalLine(int x)
{
    gdk_draw_line(pixMap, drawingGC, x, 0, x, height-1);
}

// Main drawing function:
// ---------------------
void Visualizer::draw()
{
    setColor(65535, 65535, 65535);
    gdk_draw_rectangle(pixMap, drawingGC, TRUE, 0, 0, width, height);

    drawAxes();

    for(std::list<Function*>::iterator iter = functions.begin();
        iter != functions.end(); ++iter)
    {
        Function* f = *iter;
        if(f->shouldBeDrawn() && f->isValid())
        {
            const GdkColor* c = f->getColor();
            setColor(c->red, c->green, c->blue);
            drawFunction(f);
        }
    }

    refreshArea(0, 0, width, height);
}

// Axis grid drawing function:
// --------------------------
void Visualizer::drawAxes()
{
    char number[16];

    PangoFontDescription* fontDesc =
        pango_context_get_font_description(pango_layout_get_context(pLayout));
    pango_font_description_set_size(fontDesc, 8*PANGO_SCALE);

    // Calculate the required coordinates:
    const int screenMidX = width/2;
    const double gridXStep =
        std::pow(10, std::floor(std::log10(horizontalSize/3)));
    const double yAxisLocation =
        round(-width*centerX/horizontalSize) + screenMidX;
    const int screenMidY = height/2;
    const double verticalSize = horizontalSize*aspectRatio*height/width;
    const double gridYStep =
        std::pow(10, std::floor(std::log10(aspectRatio*horizontalSize/3)));
    const double xAxisLocation =
        screenMidY - round(-height*centerY/verticalSize);

    int textHeight;
    pango_layout_get_size(pLayout, NULL, &textHeight);
    textHeight /= PANGO_SCALE;
    const int xAxisNumbersYCoord =
        xAxisLocation < 0 ? 0 :
        xAxisLocation >= height-textHeight ? height-textHeight :
        int(xAxisLocation);
    const int yAxisNumbersXCoord =
        yAxisLocation < 0 ? 0 : int(yAxisLocation);

    // Draw vertical lines:
    setColor(0x10000*3/4, 0x10000*3/4, 0x10000*3/4);

    double xCoord =
        std::ceil((centerX-horizontalSize/2)/gridXStep)*gridXStep;
    while(true)
    {
        const int screenX =
            int(round(width*(xCoord-centerX)/horizontalSize)) + screenMidX;
        if(screenX >= width) break;
        verticalLine(screenX);

        // Draw the coordinate number:
        const double rounded = round(xCoord/gridXStep)*gridXStep;
        pango_layout_set_text(pLayout, number,
                              std::sprintf(number, "%g", rounded));
        gdk_draw_layout(pixMap, drawingGC,
                        screenX+1, xAxisNumbersYCoord, pLayout);

        xCoord += gridXStep;
    }

    if(yAxisLocation >= 0 && yAxisLocation < width)
    {
        setColor(0x10000/4, 0x10000/4, 0x10000/4);
        verticalLine(int(yAxisLocation));
        setColor(0x10000*3/4, 0x10000*3/4, 0x10000*3/4);
    }

    // Draw horizontal lines:
    double yCoord =
        std::ceil((centerY-verticalSize/2)/gridYStep)*gridYStep;
    while(true)
    {
        const int screenY = screenMidY -
            int(round(height*(yCoord-centerY)/verticalSize));
        if(screenY < 0) break;
        horizontalLine(screenY);

        // Draw the coordinate number:
        if(yCoord < -1e-10 || yCoord > 1e-10)
        {
            const double rounded = round(yCoord/gridYStep)*gridYStep;
            pango_layout_set_text(pLayout, number,
                                  std::sprintf(number, "%g", rounded));
            int textWidth;
            pango_layout_get_size(pLayout, &textWidth, NULL);
            textWidth /= PANGO_SCALE;
            const int xLoc = yAxisNumbersXCoord > width-textWidth ?
                width-textWidth : yAxisNumbersXCoord+1;
            gdk_draw_layout(pixMap, drawingGC, xLoc, screenY, pLayout);
        }

        yCoord += gridYStep;
    }

    if(xAxisLocation >= 0 && xAxisLocation < height)
    {
        setColor(0x10000/4, 0x10000/4, 0x10000/4);
        horizontalLine(int(xAxisLocation));
    }
}

// Draw the given function:
// -----------------------
void Visualizer::drawFunction(Function* f)
{
    FunctionParser& fp = f->getParser();
    const int midX = width/2;
    const int midY = height/2;
    int oldY = 0;

    for(int pixelX = 0; pixelX < width; ++pixelX)
    {
        const double x = centerX + (pixelX-midX)*horizontalSize/width;
        const int pixelY = midY -
            int(round((fp.Eval(&x)-centerY)*width/aspectRatio/horizontalSize));

        if(pixelX && (pixelY < oldY-1 || pixelY > oldY+1))
        {
            if(pixelY >= 0 && pixelY < height || oldY >= 0 && oldY < height)
                gdk_draw_line(pixMap, drawingGC,
                              pixelX, pixelY, pixelX-1, oldY);
        }
        else
            gdk_draw_point(pixMap, drawingGC, pixelX, pixelY);

        oldY = pixelY;
    }
}



//---------------------------------------------------------------------
// Gtk callback functions
//---------------------------------------------------------------------

// This is called when the size of the widget changes. We need to allocate
// a new pixmap each time this happens.
gboolean Visualizer::configureCallback(GtkWidget* widget, GdkEventConfigure*,
                                       gpointer data)
{
    Visualizer* instance = reinterpret_cast<Visualizer*>(data);

    if(instance->pixMap) g_object_unref(instance->pixMap);

    instance->pixMap = gdk_pixmap_new(widget->window,
                                      widget->allocation.width,
                                      widget->allocation.height, -1);
    instance->width = widget->allocation.width;
    instance->height = widget->allocation.height;

    instance->draw();

    return FALSE;
}

// This is called when some part of the widget was hidden behind something
// and became visible. It's enough to draw that part of the pixmap.
gboolean Visualizer::exposeCallback(GtkWidget*, GdkEventExpose* event,
                                    gpointer data)
{
    Visualizer* instance = reinterpret_cast<Visualizer*>(data);
    instance->refreshArea(event->area.x, event->area.y,
                          event->area.width, event->area.height);

    return FALSE;
}

// Start the click&drag operation:
gboolean Visualizer::mouseClickCallback(GtkWidget*, GdkEventButton* button,
                                        gpointer data)
{
    Visualizer* instance = reinterpret_cast<Visualizer*>(data);
    instance->prevMouseX = int(button->x);
    instance->prevMouseY = int(button->y);
    return FALSE;
}

// Mouse motion while button is being pressed:
gboolean Visualizer::mouseMotionCallback(GtkWidget*, GdkEventMotion* motion,
                                         gpointer data)
{
    Visualizer* instance = reinterpret_cast<Visualizer*>(data);
    int newX = int(motion->x), newY = int(motion->y);

    // If button 1 and ctrl are pressed, change the aspect ratio:
    if((motion->state & GDK_BUTTON1_MASK) &&
       (motion->state & GDK_CONTROL_MASK))
    {
        const int yDiff = newY - instance->prevMouseY;
        if(yDiff > 0)
            instance->aspectRatio *= 1.0+yDiff/200.0;
        else
            instance->aspectRatio /= 1.0-yDiff/200.0;
        instance->draw();
    }
    // If button 3 is pressed, scroll:
    else if(motion->state & GDK_BUTTON3_MASK)
    {
        const int xDiff = newX - instance->prevMouseX;
        const int yDiff = newY - instance->prevMouseY;
        instance->centerX -=
            xDiff * instance->horizontalSize / instance->width;
        instance->centerY +=
            yDiff * instance->horizontalSize * instance->aspectRatio /
            instance->width;
        instance->draw();
    }
    // If button 1 is pressed, zoom:
    else if(motion->state & GDK_BUTTON1_MASK)
    {
        const int yDiff = instance->prevMouseY - newY;
        if(yDiff > 0)
            instance->horizontalSize *= 1.0+yDiff/200.0;
        else
            instance->horizontalSize /= 1.0-yDiff/200.0;
        instance->draw();
    }

    instance->prevMouseX = newX;
    instance->prevMouseY = newY;
    return FALSE;
}

// Several reset actions:
void Visualizer::resetScrollCallback(GtkToolButton*, gpointer data)
{
    Visualizer* instance = reinterpret_cast<Visualizer*>(data);

    instance->centerX = 0;
    instance->centerY = 0;
    instance->draw();
}

void Visualizer::resetZoomCallback(GtkToolButton*, gpointer data)
{
    Visualizer* instance = reinterpret_cast<Visualizer*>(data);

    instance->horizontalSize = 10;
    instance->draw();
}

void Visualizer::resetAspectRatioCallback(GtkToolButton*, gpointer data)
{
    Visualizer* instance = reinterpret_cast<Visualizer*>(data);

    instance->aspectRatio = 1;
    instance->draw();
}
